from .filecache import (
    filepaths,
    delete_cache,
    delete_files,
    create_cache,
    exists,
    delete_default,
    set_directive_function,
    remove_directive_function,
    get_cache,
    set,
)
